"""Displays currently active gpu by optimus-manager
Requires the following packages:

    * optimus-manager

"""

import core.module
import core.widget

import util.cli

class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.output))
        self.__gpumode = ""

    def output(self, _):
        return "GPU: {}".format(self.__gpumode)

    def update(self):
        cmd = "optimus-manager --print-mode"
        output = util.cli.execute(cmd).strip()

        if "intel" in output:
            self.__gpumode = "Intel"
        elif "nvidia" in output:
            self.__gpumode = "Nvidia"
        elif "amd" in output:
            self.__gpumode = "AMD"



"""
A module to show the currently active network connection (ethernet or wifi) and connection strength if the connection is wireless.

Requires the Python netifaces package and iw installed on Linux.

A simpler take on nic and network_traffic. No extra config necessary!

"""


import util.cli
import util.format

import core.module
import core.widget
import core.input

import netifaces
import socket


class Module(core.module.Module):
    @core.decorators.every(seconds=5)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.network))
        self.__is_wireless = False
        self.__is_connected = False
        self.__interface = None
        self.__message = None
        self.__signal = -110

    # Get network information to display to the user
    def network(self, widgets):
        # Determine whether there is an internet connection
        self.__is_connected = self.__attempt_connection()

        # Attempt to extract a valid network interface device
        try:
            self.__interface = netifaces.gateways()["default"][netifaces.AF_INET][1]
        except Exception:
            self.__interface = None

        # Check to see if the interface (if connected to the internet) is wireless
        if self.__is_connected and self.__interface:
            self.__is_wireless = self.__interface_is_wireless(self.__interface)

        # setup message to send to the user
        if not self.__is_connected or not self.__interface:
            self.__message = "No connection"
        elif not self.__is_wireless:
            # Assuming that if user is connected via non-wireless means that it will be ethernet
            self.__signal = -30
            self.__message = "Ethernet"
        else:
            # We have a wireless connection
            iw_dat = util.cli.execute("iwgetid")
            has_ssid = "ESSID" in iw_dat
            signal = self.__compute_signal(self.__interface)

            # If signal is None, that means that we can't compute the default interface's signal strength
            self.__signal = (
                util.format.asint(signal, minimum=-110, maximum=-30) if signal else None
            )

            ssid = (
                iw_dat[iw_dat.index(":") + 1 :].replace('"', "").strip()
                if has_ssid
                else "Unknown"
            )
            self.__message = self.__generate_wireles_message(ssid, self.__signal)

        return self.__message

    # State determined by signal strength
    def state(self, widget):
        if self.__compute_strength(self.__signal) < 50:
            return "critical"
        if self.__compute_strength(self.__signal) < 75:
            return "warning"

        return None

    # manually done for better granularity / ease of parsing strength data
    def __generate_wireles_message(self, ssid, signal):
        computed_strength = self.__compute_strength(signal)
        strength_str = str(computed_strength) if computed_strength else "?"

        return "{} {}%".format(ssid, strength_str)

    def __compute_strength(self, signal):
        return int(100 * ((signal + 100) / 70.0)) if signal else None

    # get signal strength in decibels/milliwat
    def __compute_signal(self, interface):
        # Get connection strength
        cmd = "iwconfig {}".format(interface)
        config_dat = " ".join(util.cli.execute(cmd).split())
        config_tokens = config_dat.replace("=", " ").split()

        # handle weird output
        try:
            signal = config_tokens[config_tokens.index("level") + 1]
        except Exception:
            signal = None

        return signal

    def __attempt_connection(self):
        can_connect = False
        try:
            socket.create_connection(("1.1.1.1", 53))
            can_connect = True
        except Exception:
            can_connect = False

        return can_connect

    def __interface_is_wireless(self, interface):
        is_wireless = False
        try:
            with open("/proc/net/wireless", "r") as f:
                is_wireless = interface in f.read()
                f.close()
        except Exception:
            is_wireless = False

        return is_wireless




# pylint: disable=C0111,R0903

"""Displays DNF package update information (<security>/<bugfixes>/<enhancements>/<other>)

Requires the following executable:
    * dnf

"""

import core.event
import core.module
import core.widget
import core.decorators

import util.cli


class Module(core.module.Module):
    @core.decorators.every(minutes=30)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.updates))

        self.background = True

    def updates(self, widget):
        result = []
        for t in ["security", "bugfixes", "enhancements", "other"]:
            result.append(str(widget.get(t, 0)))
        return "/".join(result)

    def update(self):
        widget = self.widget()
        res = util.cli.execute("dnf updateinfo", ignore_errors=True)

        security = 0
        bugfixes = 0
        enhancements = 0
        other = 0
        for line in res.split("\n"):
            if not line.startswith(" "):
                continue
            elif "ecurity" in line:
                for s in line.split():
                    if s.isdigit():
                        security += int(s)
            elif "ugfix" in line:
                for s in line.split():
                    if s.isdigit():
                        bugfixes += int(s)
            elif "hancement" in line:
                for s in line.split():
                    if s.isdigit():
                        enhancements += int(s)
            else:
                for s in line.split():
                    if s.isdigit():
                        other += int(s)

        widget.set("security", security)
        widget.set("bugfixes", bugfixes)
        widget.set("enhancements", enhancements)
        widget.set("other", other)


    def state(self, widget):
        cnt = 0
        for t in ["security", "bugfixes", "enhancements", "other"]:
            cnt += widget.get(t, 0)
        if cnt == 0:
            return "good"
        if cnt > 50 or widget.get("security", 0) > 0:
            return "critical"


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import pytest

from util.algorithm import *


@pytest.fixture
def someData():
    return {"a": 100, "b": 200, "c": [1, 2, 3]}


@pytest.fixture
def differentData():
    return {"x": 20, "y": "bla", "z": ["a", "b"]}


@pytest.fixture
def moreData():
    return {"n": 100}


@pytest.fixture
def overlapData():
    return {"a": 200, "c": [1, 2, 4]}


def test_merge_with_empty(someData):
    assert merge(someData, {}) == someData
    assert merge(someData, None) == None

    def test_merge_no_overwrite(someData, differentData):
        result = merge(someData, differentData)
        for k in someData:
            assert someData[k] == result[k]
        for k in self.differentData:
            assert differentData[k] == result[k]

    def test_merge_multiple(someData, differentData, moreData):
        result = merge(someData, differentData, moreData)
        for k in someData:
            assert someData[k] == result[k]
        for k in differentData:
            assert differentData[k] == result[k]
        for k in moreData:
            assert moreData[k] == result[k]

    def merge_overlap(someData, overlapData):
        result = merge(someData, overlapData)
        for k in someData:
            if not k in self.overlapData:
                assert someData[k] == result[k]
        for k in self.overlapData:
            assert overlapData[k] == result[k]


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



#!/usr/bin/env python

import os
import re
import sys
import glob


def is_psl(module):
    lib_path = os.path.dirname(os.__file__)
    old_sys = sys.path
    sys.path = [lib_path]
    is_psl = True
    try:
        __import__(module)
    except Exception as e:
        is_psl = False
    sys.path = old_sys

    return is_psl


def is_internal(module):
    if module.startswith("core.") or module == "core":
        return True
    if module.startswith("util.") or module == "util":
        return True
    if module.startswith("bumblebee_status."):
        return True
    if module.startswith("."):
        return True

    return is_psl(module)


def dependencies(filename):
    deps = []
    with open(filename) as f:
        for line in f:
            if "import" in line:
                match = re.match("\s*(from (\S+) )?import (\S+)", line)
                if not match:
                    continue
                dep = match.group(2) or match.group(3)
                if "util.popup" in dep or ("util" in line and "popup" in line):
                    deps.append("tkinter")
                if ".datetimetz" in line:
                    deps.extend(
                        dependencies("bumblebee_status/modules/contrib/datetimetz.py")
                    )
                elif not is_internal(dep):
                    deps.append(dep)
    return deps


def write_test(testname, modname, deps):
    fqmn = ".".join(["modules", testname.split(os.sep)[2], modname])
    if not os.path.exists(testname):
        with open(testname, "w") as f:
            f.writelines(
                ["import pytest\n\n",]
            )
            for dep in deps:
                f.write('pytest.importorskip("{}")\n\n'.format(dep))

    with open(testname) as f:
        for line in f:
            if "def test_load_module(" in line:
                print("skipping {}, already contains test".format(modname))
                return

    print("writing base test for {}".format(modname))
    with open(testname, "a+") as f:
        f.writelines(
            ["def test_load_module():\n", '    __import__("{}")\n\n'.format(fqmn),]
        )


def main():
    for f in glob.glob("bumblebee_status/modules/*/*.py"):
        if os.path.basename(f) == "__init__.py":
            continue

        modname = os.path.splitext(os.path.basename(f))[0]

        modpath = os.path.dirname(f)
        deps = dependencies(f)
        testname = os.path.join(
            "tests", "modules", modpath.split(os.sep)[2], "test_{}.py".format(modname)
        )

        write_test(testname, modname, deps)


if __name__ == "__main__":
    main()

# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import pytest

import util.cli
import core.config
import modules.contrib.layout_xkbswitch

def build_module():
    return modules.contrib.layout_xkbswitch.Module(
        config=core.config.Config([]),
        theme=None
    )

def test_load_module():
    __import__("modules.contrib.layout-xkbswitch")

def test_load_symbolic_link_module():
    __import__("modules.contrib.layout_xkbswitch")

def test_current_layout(mocker):
    command = mocker.patch('util.cli.execute')
    command.side_effect = ['en', 'en']

    module = build_module()
    widget = module.widget()

    module.update()

    assert widget.full_text() == 'en'

def test_current_layout_exception(mocker):
    command = mocker.patch('util.cli.execute')
    command.side_effect = RuntimeError

    module = build_module()
    widget = module.widget()

    module.update()

    assert widget.full_text() == ['n/a']

def test_input_register(mocker):
    input_register = mocker.patch('core.input.register')

    module = build_module()

    input_register.assert_called_with(
        module,
        button=core.input.LEFT_MOUSE,
        cmd=module.next_keymap
    )

def test_next_keymap(mocker):
    command = mocker.patch('util.cli.execute')

    module = build_module()
    module.next_keymap(False)

    command.assert_called_with('xkb-switch -n', ignore_errors=True)



"""Displays temperature of blugon and Controls it.

Use wheel up and down to change temperature, middle click to toggle and right click to reset temperature.

Default Values:
    * Minimum temperature: 1000 (red)
    * Maximum temperature: 20000 (blue)
    * Default temperature: 6600

Requires the following executable:
    * blugon

Parameters:
    * blugon.step: The amount of increase/decrease on scroll (default: 200)

contributed by `DTan13 <https://github.com/DTan13>`
"""

import core.module
import core.widget

import util.cli
import util.format

import os


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.full_text))
        self.__state = True
        self.__default = 6600
        self.__step = (
            util.format.asint(self.parameter("step")) if self.parameter("step") else 200
        )
        self.__max, self.__min = 20000, 1000

        file = open(os.path.expanduser("~/.config/blugon/current"))
        self.__current = int(float(file.read()))

        events = [
            {
                "type": "toggle",
                "action": self.toggle,
                "button": core.input.MIDDLE_MOUSE,
            },
            {
                "type": "blue",
                "action": self.blue,
                "button": core.input.WHEEL_UP,
            },
            {
                "type": "red",
                "action": self.red,
                "button": core.input.WHEEL_DOWN,
            },
            {
                "type": "reset",
                "action": self.reset,
                "button": core.input.RIGHT_MOUSE,
            },
        ]

        for event in events:
            core.input.register(self, button=event["button"], cmd=event["action"])

    def set_temp(self):
        temp = self.__current if self.__state else self.__default
        util.cli.execute("blugon --setcurrent={}".format(temp))

    def full_text(self, widget):
        return self.__current if self.__state else self.__default

    def state(self, widget):
        if not self.__state:
            return ["critical"]

    def toggle(self, event):
        self.__state = not self.__state
        self.set_temp()

    def reset(self, event):
        self.__current = 6600
        self.set_temp()

    def blue(self, event):
        if self.__state and (self.__current < self.__max):
            self.__current += self.__step
        self.set_temp()

    def red(self, event):
        if self.__state and (self.__current > self.__min):
            self.__current -= self.__step
        self.set_temp()


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays volume and mute status and controls for PulseAudio devices. Use wheel up and down to change volume, left click mutes, right click opens pavucontrol.

!!! This module will eventually be deprecated (since it has bad performance and high CPU load) and be replaced with "pulsectl", which is a much better drop-in replacement !!!

Aliases: pasink (use this to control output instead of input), pasource

Parameters:
    * pulseaudio.autostart: If set to 'true' (default is 'false'), automatically starts the pulseaudio daemon if it is not running
    * pulseaudio.percent_change: How much to change volume by when scrolling on the module (default is 2%)
    * pulseaudio.limit: Upper limit for setting the volume (default is 0%, which means 'no limit')
      Note: If the left and right channels have different volumes, the limit might not be reached exactly.
    * pulseaudio.showbars: 1 for showing volume bars, requires --markup=pango;
      0 for not showing volume bars (default)
    * pulseaudio.showdevicename: If set to 'true' (default is 'false'), the currently selected default device is shown.
      Per default, the sink/source name returned by "pactl list sinks short" is used as display name.

      As this name is usually not particularly nice (e.g "alsa_output.usb-Logitech_Logitech_USB_Headset-00.analog-stereo"),
      its possible to map the name to more a user friendly name.

      e.g to map "alsa_output.usb-Logitech_Logitech_USB_Headset-00.analog-stereo" to the name "Headset", add the following
      bumblebee-status config entry: pulseaudio.alsa_output.usb-Logitech_Logitech_USB_Headset-00.analog-stereo=Headset

      Furthermore its possible to specify individual (unicode) icons for all sinks/sources. e.g in order to use the icon 🎧 for the
      "alsa_output.usb-Logitech_Logitech_USB_Headset-00.analog-stereo" sink, add the following bumblebee-status config entry:
      pulseaudio.icon.alsa_output.usb-Logitech_Logitech_USB_Headset-00.analog-stereo=🎧
    * Per default a left mouse button click mutes/unmutes the device. In case you want to open a dropdown menu to change the current
      default device add the following config entry to your bumblebee-status config: pulseaudio.left-click=select_default_device_popup

Requires the following executable:
    * pulseaudio
    * pactl
    * pavucontrol
"""

import re
import logging
import functools

import core.module
import core.widget
import core.input

import util.cli
import util.graph
import util.format

try:
    import util.popup
except ImportError as e:
    logging.warning("Couldn't import util.popup: %s. Popups won't work!", e)

class Module(core.module.Module):
    def __init__(self, config, theme, channel):
        super().__init__(config, theme, core.widget.Widget(self.display))

        if util.format.asbool(self.parameter("autostart", False)):
            util.cli.execute("pulseaudio --start", ignore_errors=True)

        self._change = util.format.asint(
            self.parameter("percent_change", "2%").strip("%"), 0, 100
        )
        self._limit = util.format.asint(self.parameter("limit", "0%").strip("%"), 0)

        self._left = 0
        self._right = 0
        self._mono = 0
        self._mute = False
        self._failed = False
        self._channel = channel
        self.__selected_default_device = None
        self._showbars = util.format.asbool(self.parameter("showbars", 0))
        self.__show_device_name = util.format.asbool(
            self.parameter("showdevicename", False)
        )

        self._patterns = [
            {"expr": "Name:", "callback": (lambda line: False)},
            {
                "expr": "Mute:",
                "callback": (
                    lambda line: self.mute(False if " no" in line.lower() else True)
                ),
            },
            {"expr": "Volume:", "callback": self.getvolume},
        ]

        core.input.register(self, button=core.input.RIGHT_MOUSE, cmd="pavucontrol")

        events = [
            {"type": "mute", "action": self.toggle, "button": core.input.LEFT_MOUSE},
            {
                "type": "volume",
                "action": self.increase_volume,
                "button": core.input.WHEEL_UP,
            },
            {
                "type": "volume",
                "action": self.decrease_volume,
                "button": core.input.WHEEL_DOWN,
            },
        ]

        for event in events:
            core.input.register(self, button=event["button"], cmd=event["action"])

    def set_volume(self, amount):
        util.cli.execute(
            "pactl set-{}-{} @DEFAULT_{}@ {}".format(
                self._channel, "volume", self._channel.upper(), amount
            )
        )

    def increase_volume(self, event):
        if self._limit > 0:  # we need to check the limit
            left = int(self._left)
            right = int(self._right)
            if (
                left + self._change >= self._limit
                or right + self._change >= self._limit
            ):
                if left == right:
                    # easy case, just set to limit
                    self.set_volume("{}%".format(self._limit))
                    return
                else:
                    # don't adjust anymore, since i don't know how to update only one channel
                    return

        self.set_volume("+{}%".format(self._change))

    def decrease_volume(self, event):
        self.set_volume("-{}%".format(self._change))

    def toggle(self, event):
        util.cli.execute(
            "pactl set-{}-mute @DEFAULT_{}@ toggle".format(
                self._channel, self._channel.upper()
            )
        )

    def mute(self, value):
        self._mute = value

    def getvolume(self, line):
        if "mono" in line:
            m = re.search(r"mono:.*\s*\/\s*(\d+)%", line)
            if m:
                self._mono = m.group(1)
                self._left = 0
                self._right = 0
        else:
            m = re.search(r"left:.*\s*\/\s*(\d+)%.*right:.*\s*\/\s*(\d+)%", line)
            if m:
                self._mono = 0
                self._left = m.group(1)
                self._right = m.group(2)

    def _default_device(self):
        output = util.cli.execute("pactl info")
        pattern = "Default {}: ".format("Sink" if self._channel == "sink" else "Source")
        for line in output.split("\n"):
            if line.startswith(pattern):
                return line.replace(pattern, "")
        logging.error("no pulseaudio device found")
        return "n/a"

    def display(self, widget):
        if self._failed == True:
            return "n/a"

        vol = None
        if int(self._mono) > 0:
            vol = "{}%".format(self._mono)
            if self._showbars:
                vol = "{} {}".format(vol, util.graph.hbar(float(self._mono)))
        elif self._left == self._right:
            vol = "{}%".format(self._left)
            if self._showbars:
                vol = "{} {}".format(vol, util.graph.hbar(float(self._left)))
        else:
            vol = "{}%/{}%".format(self._left, self._right)
            if self._showbars:
                vol = "{} {}{}".format(
                    vol,
                    util.graph.hbar(float(self._left)),
                    util.graph.hbar(float(self._right)),
                )

        output = vol
        if self.__show_device_name:
            friendly_name = self.parameter(
                self.__selected_default_device, self.__selected_default_device
            )
            icon = self.parameter("icon." + self.__selected_default_device, "")
            output = (
                icon + " " + friendly_name + " | " + vol
                if icon != ""
                else friendly_name + " | " + vol
            )
        return output

    def update(self):
        try:
            self._failed = False
            channel = "sinks" if self._channel == "sink" else "sources"
            self.__selected_default_device = self._default_device()

            result = util.cli.execute("pactl list {}".format(channel))
            found = False

            for line in result.split("\n"):
                if "Name: {}".format(self.__selected_default_device) in line:
                    found = True
                    continue
                if found is False:
                    continue
                for pattern in self._patterns:
                    if not pattern["expr"] in line:
                        continue
                    if pattern["callback"](line) is False and found == True:
                        return
        except Exception as e:
            self._failed = True
            logging.exception(e)
            if util.format.asbool(self.parameter("autostart", False)):
                util.cli.execute("pulseaudio --start", ignore_errors=True)
            else:
                raise e

    def __on_sink_selected(self, sink_name):
        util.cli.execute("pactl set-default-{} {}".format(self._channel, sink_name))

    def select_default_device_popup(self, widget):
        channel = "sinks" if self._channel == "sink" else "sources"
        result = util.cli.execute("pactl list {} short".format(channel))

        menu = util.popup.menu()
        lines = result.splitlines()
        for line in lines:
            info = line.split("\t")
            try:
                friendly_name = self.parameter(info[1], info[1])
                menu.add_menuitem(
                    friendly_name,
                    callback=functools.partial(self.__on_sink_selected, info[1]),
                )
            except:
                logging.exception("Couldn't parse {}".format(channel))
                pass

        menu.show(widget)

    def state(self, widget):
        if self._mute:
            return ["warning", "muted"]
        return ["unmuted"]


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



#!/bin/bash

import sys
import json
import hashlib
import requests

rv = requests.request(
    "GET",
    "https://api.github.com/repos/tobi-wan-kenobi/bumblebee-status/releases/latest",
)

if rv.status_code != 200:
    sys.exit(1)

release = json.loads(rv.text)

tar = requests.get(f"https://github.com/tobi-wan-kenobi/bumblebee-status/archive/{release['name']}.tar.gz")
checksum = hashlib.sha512(tar.content).hexdigest()

template = ""
with open("./PKGBUILD.template") as f:
    template = f.read()

template = template.replace("<PKGVERSION>", release["name"].lstrip("v"))
template = template.replace("<SHA512SUM>", checksum)

print(template)



"""Rofication indicator

    https://github.com/DaveDavenport/Rofication
    simple module to show an icon + the number of notifications stored in rofication
    module will have normal highlighting if there are zero notifications,
                     "warning" highlighting if there are nonzero notifications,
                     "critical" highlighting if there are any critical notifications

    Parameters:
    * rofication.regolith: Switch to regolith fork of rofication, see <https://github.com/regolith-linux/regolith-rofication>.

"""

import core.module
import core.widget
import core.decorators

import sys
import socket

class Module(core.module.Module):
    @core.decorators.every(seconds=5)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.full_text))
        self.__critical = False
        self.__numnotifications = 0
        self.__regolith = self.parameter("regolith", False)


    def full_text(self, widgets):
        with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as client:
            client.connect("/tmp/rofi_notification_daemon")
            # below code will fetch two numbers in a list, e.g. ['22', '1']
            # first is total number of notifications, second is number of critical notifications
            if self.__regolith:
                client.sendall(bytes("num\n", "utf-8"))
            else:
                client.sendall(bytes("num", "utf-8"))
            val = client.recv(512)
            val = val.decode("utf-8")
            if self.__regolith:
                l = val.split(',',2)
            else:
                l = val.split('\n',2)
            self.__numnotifications = int(l[0])
            self.__critical = bool(int(l[1]))
            return self.__numnotifications

    def state(self, widget):
        # rofication doesn't really support the idea of seen vs unseen notifications
        # marking a message as "seen" actually just sets its urgency to normal
        # so, doing highlighting if any notifications are present
        if self.__critical:
            return ["critical"]
        elif self.__numnotifications:
            return ["warning"]
        return []

# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



"""Displays first upcoming event in google calendar.

Events that are set as 'all-day' will not be shown.

Requires credentials.json from a google api application where the google calendar api is installed.
On first time run the browser will open and google will ask for permission for this app to access the google calendar and then save a .gcalendar_token.json file to the credentials_path directory which stores this permission.

A refresh is done every 15 minutes.

Parameters:
    * gcalendar.time_format: Format time output. Defaults to "%H:%M".
    * gcalendar.date_format: Format date output. Defaults to "%d.%m.%y".
    * gcalendar.credentials_path: Path to credentials.json. Defaults to "~/".
    * gcalendar.locale: locale to use rather than the system default.

Requires these pip packages:
   * google-api-python-client >= 1.8.0
   * google-auth-httplib2 
   * google-auth-oauthlib
"""

# This import belongs to the google code
from __future__ import print_function

from dateutil.parser import parse as dtparse

import core.module
import core.widget
import core.decorators

import datetime
import os.path
import locale

from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError


class Module(core.module.Module):
    @core.decorators.every(minutes=15)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.first_event))
        self.__time_format = self.parameter("time_format", "%H:%M")
        self.__date_format = self.parameter("date_format", "%d.%m.%y")
        self.__credentials_path = os.path.expanduser(
            self.parameter("credentials_path", "~/")
        )
        self.__credentials = os.path.join(self.__credentials_path, "credentials.json")
        self.__token = os.path.join(self.__credentials_path, ".gcalendar_token.json")

        l = locale.getdefaultlocale()
        if not l or l == (None, None):
            l = ("en_US", "UTF-8")
        lcl = self.parameter("locale", ".".join(l))
        try:
            locale.setlocale(locale.LC_TIME, lcl.split("."))
        except Exception:
            locale.setlocale(locale.LC_TIME, ("en_US", "UTF-8"))

    def first_event(self, widget):
        SCOPES = ["https://www.googleapis.com/auth/calendar.readonly"]

        """Shows basic usage of the Google Calendar API.
        Prints the start and name of the next 10 events on the user's calendar.
        """
        creds = None
        # The file token.json stores the user's access and refresh tokens, and is
        # created automatically when the authorization flow completes for the first
        # time.
        if os.path.exists(self.__token):
            creds = Credentials.from_authorized_user_file(self.__token, SCOPES)
        # If there are no (valid) credentials available, let the user log in.
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file(
                    self.__credentials, SCOPES
                )
                creds = flow.run_local_server(port=0)
            # Save the credentials for the next run
            with open(self.__token, "w") as token:
                token.write(creds.to_json())

        try:
            service = build("calendar", "v3", credentials=creds)

            # Call the Calendar API
            now = datetime.datetime.utcnow().isoformat() + "Z"  # 'Z' indicates UTC time
            end = (
                datetime.datetime.utcnow() + datetime.timedelta(days=7)
            ).isoformat() + "Z"  # 'Z' indicates UTC time
            # Get all calendars
            calendar_list = service.calendarList().list().execute()
            event_list = []
            for calendar_list_entry in calendar_list["items"]:
                calendar_id = calendar_list_entry["id"]
                events_result = (
                    service.events()
                    .list(
                        calendarId=calendar_id,
                        timeMin=now,
                        timeMax=end,
                        singleEvents=True,
                        orderBy="startTime",
                    )
                    .execute()
                )
                events = events_result.get("items", [])

                for event in events:
                    start = dtparse(
                        event["start"].get("dateTime", event["start"].get("date"))
                    )
                    # Only add to list if not an whole day event
                    if start.tzinfo:
                        event_list.append(
                            {
                                "date": start,
                                "summary": event["summary"],
                                "type": event["eventType"],
                            }
                        )
            sorted_list = sorted(event_list, key=lambda t: t["date"])

            for gevent in sorted_list:
                if gevent["date"] >= datetime.datetime.now(datetime.timezone.utc):
                    if gevent["date"].date() == datetime.datetime.utcnow().date():
                        return str(
                            "%s %s"
                            % (
                                gevent["date"]
                                .astimezone()
                                .strftime(f"{self.__time_format}"),
                                gevent["summary"],
                            )
                        )
                    else:
                        return str(
                            "%s %s"
                            % (
                                gevent["date"]
                                .astimezone()
                                .strftime(f"{self.__date_format} {self.__time_format}"),
                                gevent["summary"],
                            )
                        )
            return "No upcoming events found."

        except:
            return None


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays information about the current song in vlc, audacious, bmp, xmms2, spotify and others

Requires the following executable:
    * playerctl

Parameters:
    * playerctl.format:   Format string (defaults to '{{artist}} - {{title}}  {{duration(position)}}/{{duration(mpris:length)}}').
      The format string is passed to 'playerctl -f' as an argument. Read `the README <https://github.com/altdesktop/playerctl#printing-properties-and-metadata>`_ for more information.
    * playerctl.layout:   Comma-separated list to change order of widgets (defaults to song, previous, pause, next)
      Widget names are: playerctl.song, playerctl.prev, playerctl.pause, playerctl.next
    * playerctl.args:     The arguments added to playerctl.
      You can check 'playerctl --help' or `its README <https://github.com/altdesktop/playerctl#using-the-cli>`_. For example, it could be '-p vlc,%any'.
    * playerctl.hide:   Hide the widgets when no players are found. Defaults to "false".

Parameters are inspired by the `spotify` module, many thanks to its developers!

contributed by `smitajit <https://github.com/smitajit>`_ - many thanks!
"""

import core.module
import core.widget
import core.input
import util.cli
import util.format

import logging

class Module(core.module.Module):
    def __init__(self, config, theme):
        super(Module, self).__init__(config, theme, [])

        self.background = True

        self.__hide = util.format.asbool(self.parameter("hide", "false"));

        self.__layout = util.format.aslist(
            self.parameter(
                "layout", "playerctl.prev, playerctl.song, playerctl.pause, playerctl.next"
            )
        )

        self.__cmd = "playerctl " + self.parameter("args", "") + " "
        self.__format = self.parameter("format", "{{artist}} - {{title}}  {{duration(position)}}/{{duration(mpris:length)}}")

        widget_map = {}
        for widget_name in self.__layout:
            widget = self.add_widget(name=widget_name)
            if widget_name == "playerctl.prev":
                widget_map[widget] = {
                    "button": core.input.LEFT_MOUSE,
                    "cmd": self.__cmd + "previous",
                }
            elif widget_name == "playerctl.pause":
                widget_map[widget] = {
                    "button": core.input.LEFT_MOUSE,
                    "cmd": self.__cmd + "play-pause",
                }
            elif widget_name == "playerctl.next":
                widget_map[widget] = {
                    "button": core.input.LEFT_MOUSE,
                    "cmd": self.__cmd + "next",
                }
            elif widget_name == "playerctl.song":
                widget_map[widget] = [
                    {
                        "button": core.input.LEFT_MOUSE,
                        "cmd": self.__cmd + "play-pause",
                    }, {
                        "button": core.input.WHEEL_UP,
                        "cmd": self.__cmd + "next",
                    }, {
                        "button": core.input.WHEEL_DOWN,
                        "cmd": self.__cmd + "previous",
                    }
                ]
            else:
                raise KeyError(
                    "The playerctl module does not have a {widget_name!r} widget".format(
                        widget_name=widget_name
                    )
                )

        for widget, callback_options in widget_map.items():
            if isinstance(callback_options, dict):
                core.input.register(widget, **callback_options)

    def hidden(self):
        return self.__hide and self.status() == None

    def status(self):
        try:
            playback_status = str(util.cli.execute(self.__cmd + "status 2>&1 || true", shell = True)).strip()
            if playback_status == "No players found":
                return None
            return playback_status
        except Exception as e:
            logging.exception(e)
            return None

    def update(self):
        playback_status = self.status()
        for widget in self.widgets():
            if playback_status:
                if widget.name == "playerctl.pause":
                    if playback_status == "Playing":
                        widget.set("state", "playing")
                    elif playback_status == "Paused":
                        widget.set("state", "paused")
                    elif playback_status == "Stopped":
                        widget.set("state", "stopped")
                    else:
                        widget.set("state", "")
                elif widget.name == "playerctl.next":
                    widget.set("state", "next")
                elif widget.name == "playerctl.prev":
                    widget.set("state", "prev")
                elif widget.name == "playerctl.song":
                    widget.full_text(self.__get_song())
            else:
                widget.set("state", "")
                widget.full_text(" ")

    def __get_song(self):
        try:
            return str(util.cli.execute(self.__cmd + "metadata -f '" + self.__format + "'")).strip()
        except Exception as e:
            logging.exception(e)
            return " "



"""Displays the current song being played and allows pausing, skipping ahead, and skipping back.

Requires the following library:
    * python-dbus

Parameters:
    * spotify.format:   Format string (defaults to '{artist} - {title}')
      Available values are: {album}, {title}, {artist}, {trackNumber}
    * spotify.layout:   Comma-separated list to change order of widgets (defaults to song, previous, pause, next)
      Widget names are: spotify.song, spotify.prev, spotify.pause, spotify.next
    * spotify.concise_controls: When enabled, allows spotify to be controlled from just the spotify.song widget.
      Concise controls are:     Left Click: Toggle Pause; Wheel Up: Next; Wheel Down; Previous.
    * spotify.bus_name: String (defaults to `spotify`)
      Available values: spotify, spotifyd

contributed by `yvesh <https://github.com/yvesh>`_ - many thanks!

added controls by `LtPeriwinkle <https://github.com/LtPeriwinkle>`_ - many thanks!

fixed icons and layout parameter by `gkeep <https://github.com/gkeep>`_ - many thanks!
"""

import sys
import dbus

import core.module
import core.widget
import core.input
import core.decorators
import util.format

import logging

class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, [])

        self.background = True

        self.__bus_name = self.parameter("bus_name", "spotify")

        self.__layout = util.format.aslist(
            self.parameter(
                "layout", "spotify.song,spotify.prev,spotify.pause,spotify.next",
            )
        )

        self.__bus = dbus.SessionBus()
        self.__song = ""
        self.__pause = ""
        self.__format = self.parameter("format", "{artist} - {title}")

        if self.__bus_name == "spotifyd":
            self.__cmd = "dbus-send --session --type=method_call --dest=org.mpris.MediaPlayer2.spotifyd \
                /org/mpris/MediaPlayer2 org.mpris.MediaPlayer2.Player."
        else:
            self.__cmd = "dbus-send --session --type=method_call --dest=org.mpris.MediaPlayer2.spotify \
                /org/mpris/MediaPlayer2 org.mpris.MediaPlayer2.Player."

        widget_map = {}
        for widget_name in self.__layout:
            widget = self.add_widget(name=widget_name)
            if widget_name == "spotify.prev":
                widget_map[widget] = {
                    "button": core.input.LEFT_MOUSE,
                    "cmd": self.__cmd + "Previous",
                }
                widget.set("state", "prev")
            elif widget_name == "spotify.pause":
                widget_map[widget] = {
                    "button": core.input.LEFT_MOUSE,
                    "cmd": self.__cmd + "PlayPause",
                }
            elif widget_name == "spotify.next":
                widget_map[widget] = {
                    "button": core.input.LEFT_MOUSE,
                    "cmd": self.__cmd + "Next",
                }
                widget.set("state", "next")
            elif widget_name == "spotify.song":
                if util.format.asbool(self.parameter("concise_controls", "false")):
                    widget_map[widget] = [
                        {
                            "button": core.input.LEFT_MOUSE,
                            "cmd": self.__cmd + "PlayPause",
                        }, {
                            "button": core.input.WHEEL_UP,
                            "cmd": self.__cmd + "Next",
                        }, {
                            "button": core.input.WHEEL_DOWN,
                            "cmd": self.__cmd + "Previous",
                        }
                    ]
            else:
                raise KeyError(
                    "The spotify module does not have a {widget_name!r} widget".format(
                        widget_name=widget_name
                    )
                )
        # is there any reason the inputs can't be directly registered above?
        for widget, callback_options in widget_map.items():
            if isinstance(callback_options, dict):
                core.input.register(widget, **callback_options)

            elif isinstance(callback_options, list): # used by concise_controls
                for opts in callback_options:
                    core.input.register(widget, **opts)


    def hidden(self):
        return self.string_song == ""

    @core.decorators.scrollable
    def __get_song(self, widget):
        bus = self.__bus
        if self.__bus_name == "spotifyd":
            spotify = bus.get_object(
                "org.mpris.MediaPlayer2.spotifyd", "/org/mpris/MediaPlayer2"
            )
        else:
            spotify = bus.get_object(
                "org.mpris.MediaPlayer2.spotify", "/org/mpris/MediaPlayer2"
            )
        spotify_iface = dbus.Interface(spotify, "org.freedesktop.DBus.Properties")
        props = spotify_iface.Get("org.mpris.MediaPlayer2.Player", "Metadata")
        self.__song = self.__format.format(
            album=str(props.get("xesam:album")),
            title=str(props.get("xesam:title")),
            artist=",".join(props.get("xesam:artist")),
            trackNumber=str(props.get("xesam:trackNumber")),
        )
        return self.__song

    def update(self):
        try:
            if self.__bus_name == "spotifyd":
                bus = self.__bus.get_object(
                    "org.mpris.MediaPlayer2.spotifyd", "/org/mpris/MediaPlayer2"
                )
            else:
                bus = self.__bus.get_object(
                    "org.mpris.MediaPlayer2.spotify", "/org/mpris/MediaPlayer2"
                )

            for widget in self.widgets():
                if widget.name == "spotify.pause":
                    playback_status = str(
                        dbus.Interface(
                            bus,
                            "org.freedesktop.DBus.Properties",
                        ).Get("org.mpris.MediaPlayer2.Player", "PlaybackStatus")
                    )
                    if playback_status == "Playing":
                        widget.set("state", "playing")
                    else:
                        widget.set("state", "paused")
                elif widget.name == "spotify.song":
                    widget.set("state", "song")
                    widget.full_text(self.__get_song(widget))

        except Exception as e:
            self.__song = ""

    @property
    def string_song(self):
        if sys.version_info.major < 3:
            return unicode(self.__song)
        return str(self.__song)



"""Shows status and load percentage of logitech's unifying device

Requires the following executable:
    * solaar (from community)

contributed by `cambid <https://github.com/cambid>`_ - many thanks!
"""

import logging

import core.module
import core.widget
import core.decorators

import util.cli


class Module(core.module.Module):
    @core.decorators.every(seconds=30)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.utilization))
        self.__battery = self.parameter("device", "")
        self.background = True
        self.__battery_status = ""
        self.__error = False
        if self.__battery != "":
            self.__cmd = f"solaar show '{self.__battery}'"
        else:
            self.__cmd = "solaar show"

    @property
    def __format(self):
        return self.parameter("format", "{}")

    def utilization(self, widget):
        return self.__format.format(self.__battery_status)

    def update(self):
        self.__error = False
        code, result = util.cli.execute(
            self.__cmd, ignore_errors=True, return_exitcode=True
        )

        if code == 0:
            for line in result.split('\n'):
                if line.count('Battery') > 0:
                    self.__battery_status = line.split(':')[1].strip()
        else:
            self.__error = True
            logging.error(f"solaar exited with {code}: {result}")

    def state(self, widget):
        if self.__error:
            return "warning"
        return "okay"


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import pytest

import core.config
import modules.contrib.dunst


def build_module():
    return modules.contrib.dunst.Module(
        config=core.config.Config([]),
        theme=None
    )


def test_load_module():
    __import__("modules.contrib.dunst")

def test_input_registration(mocker):
   input_register = mocker.patch('core.input.register')

   module = build_module()

   input_register.assert_called_with(
       module,
       button=core.input.LEFT_MOUSE,
       cmd=module.toggle_status
   )

def test_dunst_toggle(mocker):
    start_command = mocker.patch('util.cli.execute')

    module = build_module()
    start_command.assert_called_with('killall -s SIGUSR2 dunst', ignore_errors=True)

    toggle_command = mocker.patch('util.cli.execute')
    module.toggle_status(None)
    toggle_command.assert_called_with('killall -s SIGUSR1 dunst')

    widget = module.widget()
    actual = module.state(widget)
    assert actual == ['muted', 'warning']

    module.toggle_status(None)
    toggle_command.assert_called_with('killall -s SIGUSR2 dunst')

    widget = module.widget()
    actual = module.state(widget)
    assert actual == ['unmuted']

def test_dunst_toggle_exception(mocker):
    module = build_module()

    toggle_command = mocker.patch('util.cli.execute', side_effect=Exception)
    module.toggle_status(None)
    toggle_command.assert_called_with('killall -s SIGUSR1 dunst')

    widget = module.widget()
    actual = module.state(widget)
    assert actual == ['unmuted']



"""
Displays information about the public IP address associated with the default route:
    * Public IP address
    * Country Name
    * Country Code
    * City Name
    * Geographic Coordinates

Left mouse click on the widget forces immediate update.
Any change to the default route will cause the widget to update.

Requirements:
    * netifaces

Parameters:
    * publicip.format: Format string (defaults to ‘{ip} ({country_code})’)
    * Available format strings - ip, country_name, country_code, city_name, coordinates

Examples:
    * bumblebee-status -m publicip -p publicip.format="{ip} ({country_code})"
    * bumblebee-status -m publicip -p publicip.format="{ip} which is in {city_name}"
    * bumblebee-status -m publicip -p publicip.format="Your packets are right here: {coordinates}"

contributed by `tfwiii <https://github.com/tfwiii>` - many thanks!
"""

import re
import threading
import netifaces
import time

import core.module
import core.widget
import core.input
import core.decorators

import util.format
import util.location
import logging

log = logging.getLogger(__name__)


class Module(core.module.Module):
    @core.decorators.every(minutes=60)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.publicip))

        self.__previous_default_route = None
        self.__current_default_route = None
        self.background = True

        # Immediate update (override default) when left click on widget
        core.input.register(self, button=core.input.LEFT_MOUSE, cmd=self.__click_update)

        # By default show: <ip> (<2 letter country code>)
        self._format = self.parameter("format", "{ip} ({country_code})")

        self.__monitor = threading.Thread(target=self.monitor, args=())
        self.__monitor.start()

    def monitor(self):
        default_route = None
        interfaces = None
        # Initially set to True to force an info update on first pass
        information_changed = True

        self.update()

        while threading.main_thread().is_alive():
            # Look for any changes in the netifaces default route information
            try:
                current_default_route = netifaces.gateways()["default"][2]
            except:
                # error reading out default gw -> assume none exists
                current_default_route = None
            if current_default_route != default_route:
                default_route = current_default_route
                information_changed = True

            # netifaces does not check ALL routing tables which might lead to false negatives
            # (ref: http://linux-ip.net/html/routing-tables.html) so additionally... look for
            # any changes in the netifaces interfaces information which might also be an inticator
            # of a change of route/external IP
            if not information_changed: # Only check if no routing table change found
                try:
                    current_interfaces = netifaces.interfaces()
                except:
                    # error reading interfaces information -> assume none exists
                    current_interfaces = None
                if current_interfaces != interfaces:
                    interfaces = current_interfaces
                    information_changed = True

            # Update either routing or interface information has changed
            if information_changed:
                information_changed = False
                self.update()

            # Throttle the calls to netifaces
            time.sleep(1)

    def publicip(self, widget):
        if widget.get("public_ip") is None:
            return "n/a"
        return self._format.format(
            ip=widget.get("public_ip", "-"),
            country_name=widget.get("country_name", "-"),
            country_code=widget.get("country_code", "-"),
            city_name=widget.get("city_name", "-"),
            coordinates=widget.get("coordinates", "-"),
        )

    def __click_update(self, event):
        util.location.reset()

    def update(self):
        widget = self.widget()

        try:
            util.location.reset()

            # Fetch fresh location information
            __info = util.location.location_info()

            # Contstruct coordinates string
            __lat = "{:.2f}".format(__info["latitude"])
            __lon = "{:.2f}".format(__info["longitude"])
            __coords = __lat + "°N" + "," + " " + __lon + "°E"

            # Set widget values
            widget.set("public_ip", __info["public_ip"])
            widget.set("country_name", __info["country"])
            widget.set("country_code", __info["country_code"])
            widget.set("city_name", __info["city_name"])
            widget.set("coordinates", __coords)

            # Update widget values
            core.event.trigger("update", [widget.module.id], redraw_only=True)
        except Exception as ex:
            widget.set("public_ip", None)
            logging.error(str(ex))

    def state(self, widget):
        return widget.get("state", None)


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import subprocess
import shutil

def query(key):
    if shutil.which("xgetres"):
        return subprocess.run(["xgetres", key],
                capture_output=True).stdout.decode("utf-8").strip()
    else:
        raise Exception("xgetres must be installed for this theme")




import pytest

pytest.importorskip("dbus")

def test_load_module():
    __import__("modules.contrib.battery-upower")

def test_load_symbolic_link_module():
    __import__("modules.contrib.battery_upower")




"""Displays the number of pending tasks in TaskWarrior.

Requires the following library:
    * taskw

Parameters:
    * taskwarrior.taskrc : path to the taskrc file (defaults to ~/.taskrc)
    * taskwarrior.show_active: true/false(default) to show the active task ID and description when one is active, otherwise show the total number pending.


contributed by `chdorb <https://github.com/chdorb>`_ - many thanks!
"""

from taskw import TaskWarrior

import core.module
import core.widget
import core.decorators


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.output))

        self.__pending_tasks = "0"
        self.__status = "stopped"

    def update(self):
        """Return a string with the number of pending tasks from TaskWarrior
        or the descripton of an active task.

        if show.active is set in the config, show the description of the
        current active task, otherwise the number of pending tasks will be displayed.
        """
        try:
            taskrc = self.parameter("taskrc", "~/.taskrc")
            show_active = self.parameter("show_active", False)
            w = TaskWarrior(config_filename=taskrc)
            active_tasks = (
                w.filter_tasks({"start.any": "", "status": "pending"}) or None
            )
            if show_active and active_tasks:
                # this is using the first element of the list, if there happen
                # to be other active tasks, they won't be displayed.
                reporting_tasks = (
                    f"{active_tasks[0]['id']} - {active_tasks[0]['description']}"
                )
                self.__status = "active"
            else:
                reporting_tasks = len(w.filter_tasks({"status": "pending"}))
                self.__status = "stopped"
            self.__pending_tasks = reporting_tasks
        except:
            self.__pending_tasks = "n/a"
            self.__status = "stopped"

    @core.decorators.scrollable
    def output(self, _):
        """Format the task counter to output in bumblebee."""
        return "{}".format(self.__pending_tasks)

    def state(self, widget):
        """Return the set status to reflect state"""
        return self.__status


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays volume and mute status and controls for PulseAudio devices. Use wheel up and down to change volume, left click mutes, right click opens pavucontrol.

**Please prefer this module over the "pulseaudio" module, which will eventually be deprecated

Aliases: pulseout (for outputs, such as headsets, speakers), pulsein (for microphones)

NOTE: Do **not** use this module directly, but rather use either pulseout or pulsein!
NOTE2: For the parameter names below, please also use pulseout or pulsein, instead of pulsectl

Parameters:
    * pulsectl.autostart: If set to 'true' (default is 'false'), automatically starts the pulsectl daemon if it is not running
    * pulsectl.percent_change: How much to change volume by when scrolling on the module (default is 2%)
    * pulsectl.limit: Upper limit for setting the volume (default is 0%, which means 'no limit')
    * pulsectl.showbars: 'true' for showing volume bars, requires --markup=pango;
      'false' for not showing volume bars (default)
    * pulsectl.showdevicename: If set to 'true' (default is 'false'), the currently selected default device is shown.
      Per default, the sink/source name returned by "pactl list sinks short" is used as display name.

      As this name is usually not particularly nice (e.g "alsa_output.usb-Logitech_Logitech_USB_Headset-00.analog-stereo"),
      its possible to map the name to more a user friendly name.

      e.g to map "alsa_output.usb-Logitech_Logitech_USB_Headset-00.analog-stereo" to the name "Headset", add the following
      bumblebee-status config entry: pulsectl.alsa_output.usb-Logitech_Logitech_USB_Headset-00.analog-stereo=Headset

      Furthermore its possible to specify individual (unicode) icons for all sinks/sources. e.g in order to use the icon 🎧 for the
      "alsa_output.usb-Logitech_Logitech_USB_Headset-00.analog-stereo" sink, add the following bumblebee-status config entry:
      pulsectl.icon.alsa_output.usb-Logitech_Logitech_USB_Headset-00.analog-stereo=🎧
    * Per default a left mouse button click mutes/unmutes the device. In case you want to open a dropdown menu to change the current
      default device add the following config entry to your bumblebee-status config: pulsectl.left-click=select_default_device_popup

Requires the following Python module:
    * pulsectl
"""

import pulsectl

import core.module
import core.widget
import core.input
import core.event

import util.cli
import util.graph
import util.format

class Module(core.module.Module):
    def __init__(self, config, theme, type):
        super().__init__(config, theme, core.widget.Widget(self.display))
        self.background = True

        self.__type = type
        self.__volume = "n/a"
        self.__devicename = "n/a"
        self.__muted = False
        self.__showbars = util.format.asbool(self.parameter("showbars", False))
        self.__show_device_name = util.format.asbool(
            self.parameter("showdevicename", False)
        )

        self.__change = util.format.asint(
            self.parameter("percent_change", "2%").strip("%"), 0, 100
        )
        self.__limit = util.format.asint(self.parameter("limit", "0%").strip("%"), 0)

        events = [
            {
                "type": "mute",
                "action": self.toggle_mute,
                "button": core.input.LEFT_MOUSE
            },
            {
                "type": "volume",
                "action": self.increase_volume,
                "button": core.input.WHEEL_UP,
            },
            {
                "type": "volume",
                "action": self.decrease_volume,
                "button": core.input.WHEEL_DOWN,
            },
        ]

        for event in events:
            core.input.register(self, button=event["button"], cmd=event["action"])

        if util.format.asbool(self.parameter("autostart", False)):
            util.cli.execute("pulseaudio --start", ignore_errors=True)

        self.process(None)

    def display(self, _):
        res = f"{int(self.__volume*100)}%"
        if self.__showbars:
            res = f"{res} {util.graph.hbar(self.__volume*100)}"

        if self.__show_device_name:
            friendly_name = self.parameter(self.__devicename, self.__devicename)
            icon = self.parameter("icon." + self.__devicename, "")
            res = (
                icon + " " + friendly_name + " | " + res
                if icon != ""
                else friendly_name + " | " + res
            )
        return res

    def toggle_mute(self, _):
        with pulsectl.Pulse(self.id + "vol") as pulse:
            dev = self.get_device(pulse)
            pulse.mute(dev, not self.__muted)

    def change_volume(self, amount):
        with pulsectl.Pulse(self.id + "vol") as pulse:
            dev = self.get_device(pulse)
            vol = dev.volume
            vol.value_flat += amount
            if self.__limit > 0 and vol.value_flat > self.__limit/100:
                vol.value_flat = self.__limit/100
            pulse.volume_set(dev, vol)

    def increase_volume(self, _):
        self.change_volume(self.__change/100.0)

    def decrease_volume(self, _):
        self.change_volume(-self.__change/100.0)

    def get_device(self, pulse):
        devs = pulse.sink_list() if self.__type == "sink" else pulse.source_list()
        default = pulse.server_info().default_sink_name if self.__type == "sink" else pulse.server_info().default_source_name

        for dev in devs:
            if dev.name == default:
                return dev
        return devs[0] # fallback



    def process(self, _):
        with pulsectl.Pulse(self.id + "proc") as pulse:
            dev = self.get_device(pulse)
            self.__volume = dev.volume.value_flat
            self.__muted = dev.mute
            self.__devicename = dev.name
        core.event.trigger("update", [self.id], redraw_only=True)
        core.event.trigger("draw")

    def update(self):
        with pulsectl.Pulse(self.id) as pulse:
            pulse.event_mask_set(self.__type)
            pulse.event_callback_set(self.process)
            pulse.event_listen()

    def state(self, _):
        if self.__muted:
            return ["warning", "muted"]
        return ["unmuted"]

# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import pytest

import util.cli
import core.config
import modules.contrib.amixer

@pytest.fixture
def module_mock(request):
    def _module_mock(config = []):
        return modules.contrib.amixer.Module(
            config=core.config.Config(config),
            theme=None
        )

    yield _module_mock

@pytest.fixture
def amixer_mock():
    def _mock(device='Master', volume='10%', state='on'):
        return """
        Simple mixer control '{device}',0
          Capabilities: pvolume pswitch pswitch-joined
          Playback channels: Front Left - Front Right
          Limits: Playback 0 - 65536
          Mono:
          Front Left: Playback 55705 [{volume}%] [{state}]
          Front Right: Playback 55705 [{volume}%] [{state}]
        """.format(
            device=device,
            volume=volume,
            state=state
        )

    return _mock

def test_load_module():
    __import__("modules.contrib.amixer")

def test_initial_full_text(module_mock, amixer_mock, mocker):
    module = module_mock()
    assert module.widget().full_text() == 'n/a'

def test_input_registration(mocker):
   input_register = mocker.patch('core.input.register')

   module = modules.contrib.amixer.Module(
       config=core.config.Config([]),
       theme=None
   )

   input_register.assert_any_call(
       module,
       button=core.input.WHEEL_DOWN,
       cmd=module.decrease_volume
   )

   input_register.assert_any_call(
       module,
       button=core.input.WHEEL_UP,
       cmd=module.increase_volume
   )

   input_register.assert_any_call(
       module,
       button=core.input.LEFT_MOUSE,
       cmd=module.toggle
   )

def test_volume_update(module_mock, amixer_mock, mocker):
    mocker.patch(
        'util.cli.execute',
        return_value=amixer_mock(volume='25%', state='on')
    )

    module = module_mock()
    widget = module.widget()

    module.update()
    assert widget.full_text() == '25%'
    assert module.state(widget) == ['unmuted']

def test_muted_update(module_mock, amixer_mock, mocker):
    mocker.patch(
        'util.cli.execute',
        return_value=amixer_mock(volume='50%', state='off')
    )

    module = module_mock()
    widget = module.widget()

    module.update()
    assert widget.full_text() == '50%'
    assert module.state(widget) == ['warning', 'muted']

def test_exception_update(module_mock, mocker):
    mocker.patch(
        'util.cli.execute',
        side_effect=Exception
    )

    module = module_mock()
    widget = module.widget()

    module.update()
    assert widget.full_text() == 'n/a'

def test_unavailable_amixer(module_mock, mocker):
    mocker.patch('util.cli.execute', return_value='Invalid')

    module = module_mock()
    widget = module.widget()

    module.update()
    assert widget.full_text() == '0%'

def test_toggle(module_mock, mocker):
    command = mocker.patch('util.cli.execute')
    module = module_mock()
    module.toggle(False)
    command.assert_called_once_with('amixer -q set Master,0 toggle')

def test_default_volume(module_mock, mocker):
    module = module_mock()

    command = mocker.patch('util.cli.execute')
    module.increase_volume(False)
    command.assert_called_once_with('amixer -q set Master,0 4%+')

    command = mocker.patch('util.cli.execute')
    module.decrease_volume(False)
    command.assert_called_once_with('amixer -q set Master,0 4%-')

def test_custom_volume(module_mock, mocker):
    module = module_mock(['-p', 'amixer.percent_change=25'])

    command = mocker.patch('util.cli.execute')
    module.increase_volume(False)
    command.assert_called_once_with('amixer -q set Master,0 25%+')

    command = mocker.patch('util.cli.execute')
    module.decrease_volume(False)
    command.assert_called_once_with('amixer -q set Master,0 25%-')

def test_custom_device(module_mock, mocker):
    mocker.patch('util.cli.execute')
    module = module_mock(['-p', 'amixer.device=CustomMaster'])

    command = mocker.patch('util.cli.execute')
    module.toggle(False)
    command.assert_called_once_with('amixer -q set CustomMaster toggle')

    command = mocker.patch('util.cli.execute')
    module.increase_volume(False)
    command.assert_called_once_with('amixer -q set CustomMaster 4%+')

    command = mocker.patch('util.cli.execute')
    module.decrease_volume(False)
    command.assert_called_once_with('amixer -q set CustomMaster 4%-')




import pytest
from unittest import TestCase, mock

import core.config
import core.widget
import modules.core.memory

def build_module(args = []):
    config = core.config.Config(args)
    return modules.core.memory.Module(config=config, theme=None)

def memory_widget(module):
    return module.widgets()[0]

def meminfo_mock(
    total,
    available,
    unit = 'kB',
    free = 0,
    buffers = 0,
    cached = 0,
    slab = 0
):
    data = []
    states = [
        ('MemTotal', total),
        ('MemAvailable', available),
        ('MemFree', free),
        ('Buffers', buffers),
        ('Cached', cached),
        ('Slab', slab)
    ]

    for i, (key, value) in enumerate(states):
        data.append('{}: {} {}'.format(key, value, unit))

    return '\n'.join(data)

class TestMemory(TestCase):
    def test_load_module(self):
        __import__("modules.core.memory")

    @mock.patch('builtins.open', mock.mock_open(read_data=meminfo_mock(2048, 1024)))
    def test_default_healthy_state(self):
        module = build_module()
        module.update()

        widget = memory_widget(module)

        assert widget.full_text() == '1.00MiB/2.00MiB (50.00%)'
        assert module.state(widget) == None

    @mock.patch('builtins.open', mock.mock_open(read_data=meminfo_mock(8196, 1024)))
    def test_default_warning_state(self):
        module = build_module()
        module.update()

        widget = memory_widget(module)

        assert widget.full_text() == '7.00MiB/8.00MiB (87.51%)'
        assert module.state(widget) == 'warning'

    @mock.patch('builtins.open', mock.mock_open(read_data=meminfo_mock(2048, 0)))
    def test_default_critical_state(self):
        module = build_module()
        module.update()

        widget = memory_widget(module)

        assert widget.full_text() == '2.00MiB/2.00MiB (100.00%)'
        assert module.state(widget) == 'critical'

    @mock.patch('builtins.open', mock.mock_open(read_data=meminfo_mock(4096, 3068)))
    def test_custom_warning_parameter(self):
        module = build_module(['-p', 'memory.warning=20'])
        module.update()

        widget = memory_widget(module)

        assert widget.full_text() == '1.00MiB/4.00MiB (25.10%)'
        assert module.state(widget) == 'warning'

    @mock.patch('builtins.open', mock.mock_open(read_data=meminfo_mock(8196, 4096)))
    def test_custom_critical_parameter(self):
        module = build_module(['-p', 'memory.critical=50'])
        module.update()

        widget = memory_widget(module)

        assert widget.full_text() == '4.00MiB/8.00MiB (50.02%)'
        assert module.state(widget) == 'critical'

    @mock.patch('builtins.open', mock.mock_open(read_data=meminfo_mock(2048, 1024)))
    def test_usedonly_parameter(self):
        module = build_module(['-p', 'memory.usedonly=true'])
        module.update()

        widget = memory_widget(module)

        assert widget.full_text() == '1.00MiB'
        assert module.state(widget) == None

    @mock.patch('builtins.open', mock.mock_open(read_data=meminfo_mock(2048, 1024)))
    def test_format_parameter(self):
        module = build_module(['-p', 'memory.format={used}.{total}'])
        module.update()

        widget = memory_widget(module)

        assert widget.full_text() == '1.00MiB.2.00MiB'
        assert module.state(widget) == None

    @mock.patch('builtins.open', mock.mock_open(read_data=meminfo_mock(2048, 1024)))
    def test_format_parameter_with_percent(self):
        module = build_module(['-p', 'memory.format={percent}%'])
        module.update()

        widget = memory_widget(module)

        assert widget.full_text() == '50.0%'
        assert module.state(widget) == None


    @mock.patch('builtins.open', mock.mock_open(read_data=meminfo_mock(8196, 4096, 'mB')))
    def test_mb_unit(self):
        module = build_module()
        module.update()

        widget = memory_widget(module)

        assert widget.full_text() == '4.00GiB/8.00GiB (50.02%)'
        assert module.state(widget) == None

    @mock.patch('builtins.open', mock.mock_open(read_data=meminfo_mock(2, 1, 'gB')))
    def test_gb_unit(self):
        module = build_module()
        module.update()

        widget = memory_widget(module)

        assert widget.full_text() == '1.00GiB/2.00GiB (50.00%)'
        assert module.state(widget) == None




"""Display information about the currently running emerge process.

Requires the following executable:
    * emerge

Parameters:
    * emerge_status.format: Format string (defaults to '{current}/{total} {action} {category}/{pkg}')

This code is based on emerge_status module from p3status [1] original created by AnwariasEu.

[1] https://github.com/ultrabug/py3status/blob/master/py3status/modules/emerge_status.py 
"""

import re
import copy

import core.module
import core.widget
import core.decorators

import util.cli
import util.format


class Module(core.module.Module):
    @core.decorators.every(seconds=10)
    def __init__(self, config, theme):
        super().__init__(config, theme, [])
        self.__format = self.parameter(
            "format", "{current}/{total} {action} {category}/{pkg}"
        )
        self.__ret_default = {
            "action": "",
            "category": "",
            "current": 0,
            "pkg": "",
            "total": 0,
        }

    def update(self):
        response = {}
        ret = copy.deepcopy(self.__ret_default)
        if self.__emerge_running():
            ret = self.__get_progress()

            widget = self.widget("status")
            if not widget:
                widget = self.add_widget(name="status")

            if ret["total"] == 0:
                widget.full_text("emrg calculating...")
            else:
                widget.full_text(
                    " ".join(
                        self.__format.format(
                            current=ret["current"],
                            total=ret["total"],
                            action=ret["action"],
                            category=ret["category"],
                            pkg=ret["pkg"],
                        ).split()
                    )
                )
        else:
            self.clear_widgets()

    def __emerge_running(self):
        """
        Check if emerge is running.
        Returns true if at least one instance of emerge is running.
        """
        try:
            util.cli.execute("pgrep emerge")
            return True
        except Exception:
            return False

    def __get_progress(self):
        """
        Get current progress of emerge.
        Returns a dict containing current and total value.
        """
        input_data = []
        ret = {}

        # traverse emerge.log from bottom up to get latest information
        last_lines = util.cli.execute("tail -50 /var/log/emerge.log")
        input_data = last_lines.split("\n")
        input_data.reverse()

        for line in input_data:
            if "*** terminating." in line:
                # copy content of ret_default, not only the references
                ret = copy.deepcopy(self.__ret_default)
                break
            else:
                status_re = re.compile(
                    r"\((?P<cu>[\d]+) of (?P<t>[\d]+)\) "
                    r"(?P<a>[a-zA-Z/]+( [a-zA-Z]+)?) "
                    r"\((?P<ca>[\w\-]+)/(?P<p>[\w.]+)"
                )
                res = status_re.search(line)
                if res is not None:
                    ret["action"] = res.group("a").lower()
                    ret["category"] = res.group("ca")
                    ret["current"] = res.group("cu")
                    ret["pkg"] = res.group("p")
                    ret["total"] = res.group("t")
                    break
        return ret


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays the current color temperature of redshift

Requires the following executable:
    * redshift

Parameters:
    * redshift.location : location provider, either of 'auto' (default), 'geoclue2',
      'ipinfo' or 'manual'
      'auto' uses whatever redshift is configured to do
    * redshift.lat : latitude if location is set to 'manual'
    * redshift.lon : longitude if location is set to 'manual'
    * redshift.show_transition: information about the transitions (x% day) defaults to True
    * redshift.adjust: set this to 'true' (defaults to false) to let bumblebee-status adjust color temperature, instead of just showing the current settings
"""

import re
import threading

import core.module
import core.widget
import core.input
import core.decorators

import util.cli
import util.format
import util.location


def get_redshift_value(module):
    widget = module.widget()
    location = module.parameter("location", "auto")
    lat = module.parameter("lat", None)
    lon = module.parameter("lon", None)

    # Even if location method is set to manual, if we have no lat or lon,
    # fall back to the geoclue2 method.
    if location == "manual" and (lat is None or lon is None):
        location = "geoclue2"

    command = ["redshift"]

    if util.format.asbool(module.parameter("adjust", "false")) == True:
        command.extend(["-o", "-v"])
    else:
        command.append("-p")

    if location == "manual":
        command.extend(["-l", "{}:{}".format(lat, lon)])
    if location == "geoclue2":
        command.extend(["-l", "geoclue2"])

    try:
        res = util.cli.execute(" ".join(command))
    except Exception:
        res = ""
    widget.set("temp", "n/a")
    widget.set("transition", "")
    widget.set("state", "day")
    for line in res.split("\n"):
        line = line.lower()
        if "temperature" in line:
            widget.set("temp", line.split(" ")[2].upper())
        if "period" in line:
            state = line.split(" ")[1]
            if "day" in state:
                widget.set("state", "day")
            elif "night" in state:
                widget.set("state", "night")
            else:
                widget.set("state", "transition")
                match = re.search(r"(\d+)\.\d+% ([a-z]+)", line)
                widget.set(
                    "transition", "({}% {})".format(match.group(1), match.group(2))
                )
    core.event.trigger("update", [widget.module.id], redraw_only=True)


class Module(core.module.Module):
    @core.decorators.every(seconds=10)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.text))

        self.__thread = None
        self.show_transition = util.format.asbool(
            self.parameter("show_transition", True)
        )

        if self.parameter("location", "") == "ipinfo":
            # override lon/lat with ipinfo
            try:
                location = util.location.coordinates()
                self.set("lat", location[0])
                self.set("lon", location[1])
                self.set("location", "manual")
            except Exception:
                # Fall back to geoclue2.
                self.set("location", "geoclue2")

        self._text = ""

    def text(self, widget):
        val = widget.get("temp", "n/a")
        transition = widget.get("transition", "")
        if transition and self.show_transition:
            val = "{} {}".format(val, transition)
        return val

    def update(self):
        if self.__thread is not None and self.__thread.is_alive():
            return
        self.__thread = threading.Thread(target=get_redshift_value, args=(self,))
        self.__thread.start()

    def state(self, widget):
        return widget.get("state", None)


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import sys

from core.input import trigger, LEFT_MOUSE, RIGHT_MOUSE
from core.config import Config
from modules.core.xrandr import Module


MOCK_EXECUTE = "modules.core.xrandr.util.cli.execute"


def mock_xrandr(mocker, xrandr_output):
    return mocker.patch(MOCK_EXECUTE, return_value=xrandr_output)


def assert_widgets(module, *expected_widgets):
    assert len(module.widgets()) == len(expected_widgets)

    for widget, (name, state, pos) in zip(module.widgets(), expected_widgets):
        assert widget.name == name
        assert widget.get("state") == state
        assert widget.get("pos") == pos


def assert_trigger(xrandr_cli, module, widget_index, button, expected_command):
    xrandr_cli.reset_mock()

    widget = module.widgets()[widget_index]
    trigger({"button": button, "instance": widget.id, "name": module.id})

    if expected_command is None:
        xrandr_cli.assert_called_once_with("xrandr -q")
    else:
        assert xrandr_cli.call_count == 2
        xrandr_cli.assert_any_call(expected_command)
        xrandr_cli.assert_called_with("xrandr -q")


def test_autoupdate(mocker):
    xrandr_cli = mock_xrandr(mocker, HDMI_CONNECTED_ACTIVE)
    module = Module(Config([]), theme=None)
    module.update()
    xrandr_cli.assert_called_once_with("xrandr -q")

    assert_widgets(module, ("eDP-1-1", "on", 0), ("HDMI-1-1", "on", 1920))

    assert_trigger(xrandr_cli, module, 0, LEFT_MOUSE, "xrandr --output eDP-1-1 --off")
    assert_trigger(xrandr_cli, module, 0, RIGHT_MOUSE, "xrandr --output eDP-1-1 --off")
    assert_trigger(xrandr_cli, module, 1, LEFT_MOUSE, "xrandr --output HDMI-1-1 --off")
    assert_trigger(xrandr_cli, module, 1, RIGHT_MOUSE, "xrandr --output HDMI-1-1 --off")


def test_display_off(mocker):
    xrandr_cli = mock_xrandr(mocker, HDMI_CONNECTED_INACTIVE)
    module = Module(Config([]), theme=None)
    module.update()
    xrandr_cli.assert_called_once_with("xrandr -q")

    assert_widgets(module, ("eDP-1-1", "on", 0), ("HDMI-1-1", "off", sys.maxsize))

    assert_trigger(xrandr_cli, module, 0, LEFT_MOUSE, None)
    assert_trigger(xrandr_cli, module, 0, RIGHT_MOUSE, None)
    assert_trigger(xrandr_cli, module, 1, LEFT_MOUSE, "xrandr --output HDMI-1-1 --auto --left-of eDP-1-1")
    assert_trigger(xrandr_cli, module, 1, RIGHT_MOUSE, "xrandr --output HDMI-1-1 --auto --right-of eDP-1-1")


def test_no_autoupdate(mocker):
    xrandr_cli = mock_xrandr(mocker, HDMI_CONNECTED_ACTIVE)
    module = Module(Config(["-p", "xrandr.autoupdate=false"]), theme=None)
    module.update()
    xrandr_cli.assert_called_once_with("xrandr -q")

    assert_widgets(
        module, ("eDP-1-1", "on", 0), ("HDMI-1-1", "on", 1920), (None, "refresh", None)
    )

    assert_trigger(xrandr_cli, module, 0, LEFT_MOUSE, "xrandr --output eDP-1-1 --off")
    assert_trigger(xrandr_cli, module, 0, RIGHT_MOUSE, "xrandr --output eDP-1-1 --off")
    assert_trigger(xrandr_cli, module, 1, LEFT_MOUSE, "xrandr --output HDMI-1-1 --off")
    assert_trigger(xrandr_cli, module, 1, RIGHT_MOUSE, "xrandr --output HDMI-1-1 --off")


def test_exclude(mocker):
    xrandr_cli = mock_xrandr(mocker, HDMI_CONNECTED_ACTIVE)
    module = Module(Config(["-p", "xrandr.exclude=eDP"]), theme=None)
    module.update()
    xrandr_cli.assert_called_once_with("xrandr -q")

    assert_widgets(module, ("HDMI-1-1", "on", 1920))

    assert_trigger(xrandr_cli, module, 0, LEFT_MOUSE, "xrandr --output HDMI-1-1 --off")
    assert_trigger(xrandr_cli, module, 0, RIGHT_MOUSE, "xrandr --output HDMI-1-1 --off")


def test_exclude_off(mocker):
    xrandr_cli = mock_xrandr(mocker, HDMI_CONNECTED_INACTIVE)
    module = Module(Config(["-p", "xrandr.exclude=eDP"]), theme=None)
    module.update()
    xrandr_cli.assert_called_once_with("xrandr -q")

    assert_widgets(module, ("HDMI-1-1", "off", sys.maxsize))

    assert_trigger(xrandr_cli, module, 0, LEFT_MOUSE, "xrandr --output HDMI-1-1 --auto --left-of eDP-1-1")
    assert_trigger(xrandr_cli, module, 0, RIGHT_MOUSE, "xrandr --output HDMI-1-1 --auto --right-of eDP-1-1")


def test_no_autotoggle_inactive_connected(mocker):
    xrandr_cli = mock_xrandr(mocker, HDMI_DISCONNECTED_INACTIVE)
    module = Module(Config([]), theme=None)
    module.update()
    xrandr_cli.assert_called_once_with("xrandr -q")

    assert_widgets(module, ("eDP-1-1", "on", 0))

    xrandr_cli.return_value = HDMI_CONNECTED_INACTIVE
    xrandr_cli.reset_mock()

    module.update()
    xrandr_cli.assert_called_once_with("xrandr -q")


def test_no_autotoggle_active_disconnected(mocker):
    xrandr_cli = mock_xrandr(mocker, HDMI_CONNECTED_ACTIVE)
    module = Module(Config([]), theme=None)
    module.update()
    xrandr_cli.assert_called_once_with("xrandr -q")

    assert_widgets(module, ("eDP-1-1", "on", 0), ("HDMI-1-1", "on", 1920))

    xrandr_cli.return_value = HDMI_DISCONNECTED_ACTIVE
    xrandr_cli.reset_mock()

    module.update()
    xrandr_cli.assert_called_once_with("xrandr -q")


def test_autotoggle_excluded_inactive_connected(mocker):
    xrandr_cli = mock_xrandr(mocker, HDMI_DISCONNECTED_INACTIVE)
    module = Module(Config(["-p", "xrandr.autotoggle=true", "xrandr.exclude=HDMI"]), theme=None)
    module.update()
    xrandr_cli.assert_called_once_with("xrandr -q")

    assert_widgets(module, ("eDP-1-1", "on", 0))

    xrandr_cli.return_value = HDMI_CONNECTED_INACTIVE
    xrandr_cli.reset_mock()

    module.update()
    xrandr_cli.assert_called_once_with("xrandr -q")


def test_autotoggle_excluded_active_disconnected(mocker):
    xrandr_cli = mock_xrandr(mocker, HDMI_CONNECTED_ACTIVE)
    module = Module(Config(["-p", "xrandr.autotoggle=true", "xrandr.exclude=HDMI"]), theme=None)
    module.update()
    xrandr_cli.assert_called_once_with("xrandr -q")

    assert_widgets(module, ("eDP-1-1", "on", 0))

    xrandr_cli.return_value = HDMI_DISCONNECTED_ACTIVE
    xrandr_cli.reset_mock()

    module.update()
    xrandr_cli.assert_called_once_with("xrandr -q")


def test_autotoggle_active_disconnected(mocker):
    xrandr_cli = mock_xrandr(mocker, HDMI_CONNECTED_ACTIVE)
    module = Module(Config(["-p", "xrandr.autotoggle=true"]), theme=None)
    module.update()
    xrandr_cli.assert_called_once_with("xrandr -q")

    assert_widgets(module, ("eDP-1-1", "on", 0), ("HDMI-1-1", "on", 1920))

    xrandr_cli.return_value = HDMI_DISCONNECTED_ACTIVE
    xrandr_cli.reset_mock()

    module.update()
    assert xrandr_cli.call_count == 2
    xrandr_cli.assert_any_call("xrandr -q")
    xrandr_cli.assert_called_with("xrandr --output HDMI-1-1 --off")


def test_autotoggle_inactive_disconnected(mocker):
    xrandr_cli = mock_xrandr(mocker, HDMI_CONNECTED_INACTIVE)
    module = Module(Config(["-p", "xrandr.autotoggle=true"]), theme=None)
    module.update()
    xrandr_cli.assert_called_once_with("xrandr -q")

    assert_widgets(module, ("eDP-1-1", "on", 0), ("HDMI-1-1", "off", sys.maxsize))

    xrandr_cli.return_value = HDMI_DISCONNECTED_INACTIVE
    xrandr_cli.reset_mock()

    module.update()
    xrandr_cli.assert_called_once_with("xrandr -q")


def test_autotoggle_active_connected(mocker):
    xrandr_cli = mock_xrandr(mocker, HDMI_DISCONNECTED_ACTIVE)
    module = Module(Config(["-p", "xrandr.autotoggle=true"]), theme=None)
    module.update()
    xrandr_cli.assert_called_once_with("xrandr -q")

    assert_widgets(module, ("eDP-1-1", "on", 0))

    xrandr_cli.return_value = HDMI_CONNECTED_ACTIVE
    xrandr_cli.reset_mock()

    module.update()
    xrandr_cli.assert_called_once_with("xrandr -q")


def test_autotoggle_inactive_connected(mocker):
    xrandr_cli = mock_xrandr(mocker, HDMI_DISCONNECTED_INACTIVE)
    module = Module(Config(["-p", "xrandr.autotoggle=true"]), theme=None)
    module.update()
    xrandr_cli.assert_called_once_with("xrandr -q")

    assert_widgets(module, ("eDP-1-1", "on", 0))

    xrandr_cli.return_value = HDMI_CONNECTED_INACTIVE
    xrandr_cli.reset_mock()

    module.update()
    assert xrandr_cli.call_count == 2
    xrandr_cli.assert_any_call("xrandr -q")
    xrandr_cli.assert_called_with("xrandr --output HDMI-1-1 --auto --right-of eDP-1-1")


def test_autotoggle_left(mocker):
    xrandr_cli = mock_xrandr(mocker, HDMI_DISCONNECTED_INACTIVE)
    module = Module(Config(["-p", "xrandr.autotoggle=true", "xrandr.autotoggle_side=left"]), theme=None)
    module.update()
    xrandr_cli.assert_called_once_with("xrandr -q")

    assert_widgets(module, ("eDP-1-1", "on", 0))

    xrandr_cli.return_value = HDMI_CONNECTED_INACTIVE
    xrandr_cli.reset_mock()

    module.update()
    assert xrandr_cli.call_count == 2
    xrandr_cli.assert_any_call("xrandr -q")
    xrandr_cli.assert_called_with("xrandr --output HDMI-1-1 --auto --left-of eDP-1-1")


# xrandr sample data

HDMI_CONNECTED_ACTIVE = """
Screen 0: minimum 8 x 8, current 4480 x 1440, maximum 32767 x 32767
eDP-1-1 connected primary 1920x1080+0+0 344mm x 193mm
   1920x1080     60.00*+  59.93    48.00
   1680x1050     59.95    59.88
   1600x1024     60.17
   1400x1050     59.98
   1280x1024     60.02
   1440x900      59.89
   1280x960      60.00
   1360x768      59.80    59.96
   1152x864      60.00
   1024x768      60.04    60.00
   960x720       60.00
   928x696       60.05
   896x672       60.01
   960x600       60.00
   960x540       59.99
   800x600       60.00    60.32    56.25
   840x525       60.01    59.88
   800x512       60.17
   700x525       59.98
   640x512       60.02
   720x450       59.89
   640x480       60.00    59.94
   680x384       59.80    59.96
   576x432       60.06
   512x384       60.00
   400x300       60.32    56.34
   320x240       60.05
HDMI-1-1 connected 2560x1440+1920+0 596mm x 335mm
   2560x1440     59.95*+
   1920x1080     60.00    50.00    59.94
   1920x1080i    60.00    50.00    59.94
   1680x1050     59.88
   1600x900      60.00
   1280x1024     60.02
   1280x800      59.91
   1280x720      60.00    50.00    59.94
   1024x768      60.00
   800x600       60.32
   720x576       50.00
   720x576i      50.00
   720x480       60.00    59.94
   720x480i      60.00    59.94
   640x480       60.00    59.94
"""


HDMI_CONNECTED_INACTIVE = """
eDP-1-1 connected primary 1920x1080+0+0 344mm x 193mm
   1920x1080     60.00*+  59.93    48.00
HDMI-1-1 connected
   2560x1440     59.95 +
"""

HDMI_DISCONNECTED_ACTIVE = """
eDP-1-1 connected primary 1920x1080+0+0 344mm x 193mm
   1920x1080     60.00*+  59.93    48.00
HDMI-1-1 disconnected 2560x1440+1920+0 0mm x 0mm
  2560x1440 (0x6b) 241.500MHz +HSync +VSync
        h: width  2560 start 2608 end 2640 total 2720 skew    0 clock  88.79KHz
        v: height 1440 start 1442 end 1447 total 1481           clock  59.95Hz
"""

HDMI_DISCONNECTED_INACTIVE = """
eDP-1-1 connected primary 1920x1080+0+0 344mm x 193mm
   1920x1080     60.00*+  59.93    48.00
HDMI-1-1 disconnected
"""



import pytest

import util.cli
import core.config
import modules.contrib.solaar


@pytest.fixture
def module():
    module = modules.contrib.solaar.Module(
        config=core.config.Config([]),
        theme=None
    )

    yield module


def test_load_module():
    __import__("modules.contrib.solaar")


def test_with_unknown_code(module, mocker):
    mocker.patch('util.cli.execute', return_value=(99, 'error'))
    logger = mocker.patch('logging.error')

    module.update()

    logger.assert_called_with('solaar exited with {}: {}'.format(99, 'error'))

    widget = module.widget()
    assert module.state(widget) == 'warning'
    assert module.hidden() == False



"""Displays the status of Gentoo portage operations.

Parameters:
    * portage_status.logfile: logfile for portage (default is /var/log/emerge.log)

contributed by `andrewreisner <https://github.com/andrewreisner>`_ - many thanks!
"""

import os

import core.module
import core.widget


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.output))
        self.__logfile = self.parameter("logfile", "/var/log/emerge.log")
        self.clear()

    def clear(self):
        self.__action = ""
        self.__package = ""
        self.__status = ""

    def output(self, widget):
        return " ".join(
            [
                atom
                for atom in (self.__action, self.__package, self.__status)
                if atom != ""
            ]
        )

    def state(self, widgets):
        if self.__action == "":
            return "idle"
        return "active"

    def update(self):
        try:
            with open(self.__logfile, "rb") as f:
                f.seek(-2, os.SEEK_END)
                while f.read(1) != b"\n":
                    f.seek(-2, os.SEEK_CUR)
                last_line = f.readline().decode()
                if "===" in last_line:
                    if "Unmerging..." in last_line:
                        self.__action = "Unmerging"
                        package_beg = last_line.find("(") + 1
                        package_end = last_line.find("-", last_line.find("/")) - 1
                        self.__package = last_line[package_beg : package_end + 1]
                    else:  # merging
                        status_beg = last_line.find("(")
                        status_end = last_line.find(")")
                        self.__status = last_line[status_beg : status_end + 1]
                        package_beg = last_line.find("(", status_end) + 1
                        package_end = (
                            package_beg
                            + last_line[package_beg:].find(
                                "-", last_line[package_beg:].find("/")
                            )
                            - 1
                        )
                        self.__package = last_line[package_beg : package_end + 1]
                        action_beg = status_end + 2
                        action_end = package_beg - 3
                        self.__action = last_line[action_beg : action_end + 1]
                else:
                    self.clear()
        except Exception:
            self.clear()



import pytest
from unittest import TestCase, mock

import core.config
import core.widget
import modules.core.cpu

pytest.importorskip("psutil")

def build_module(percpu=False):
    config = core.config.Config(["-p", "percpu={}".format(percpu)])
    config.set("cpu.percpu", percpu)
    return modules.core.cpu.Module(config=config, theme=None)

def cpu_widget(module):
    return module.widgets()[0]

class TestCPU(TestCase):
    def test_load_module(self):
        __import__("modules.core.cpu")

    @mock.patch('psutil.cpu_percent')
    def test_cpu_percent(self, cpu_percent_mock):
        cpu_percent_mock.return_value = 5
        module = build_module()

        assert cpu_widget(module).full_text() == '5.0%'

    @mock.patch('psutil.cpu_percent')
    def test_cpu_percent_update(self, cpu_percent_mock):
        cpu_percent_mock.return_value = 10
        module = build_module()

        assert cpu_widget(module).full_text() == '10.0%'

        cpu_percent_mock.return_value = 20
        module.update()

        assert cpu_widget(module).full_text() == '20.0%'

    @mock.patch('psutil.cpu_percent')
    def test_healthy_state(self, cpu_percent_mock):
        cpu_percent_mock.return_value = 50
        module = build_module()

        assert module.state(module.widget()) == None

    @mock.patch('psutil.cpu_percent')
    def test_warning_state(self, cpu_percent_mock):
        cpu_percent_mock.return_value = 75
        module = build_module()

        assert module.state(module.widget()) == 'warning'

    @mock.patch('psutil.cpu_percent')
    def test_critical_state(self, cpu_percent_mock):
        cpu_percent_mock.return_value = 82
        module = build_module()

        assert module.state(module.widget()) == 'critical'

    @mock.patch('psutil.cpu_percent')
    def test_healthy_state_percpu(self, cpu_percent_mock):
        cpu_percent_mock.return_value = [50,42,47]
        module = build_module(percpu=True)

        for widget in module.widgets():
            assert module.state(widget) == None

    @mock.patch('psutil.cpu_percent')
    def test_warning_state_percpu(self, cpu_percent_mock):
        cpu_percent_mock.return_value = [50,72,47]
        module = build_module(percpu=True)

        assert module.state(module.widgets()[0]) == None
        assert module.state(module.widgets()[1]) == "warning"
        assert module.state(module.widgets()[2]) == None

    @mock.patch('psutil.cpu_percent')
    def test_warning_state_percpu(self, cpu_percent_mock):
        cpu_percent_mock.return_value = [50,72,99]
        module = build_module(percpu=True)

        assert module.state(module.widgets()[0]) == None
        assert module.state(module.widgets()[1]) == "warning"
        assert module.state(module.widgets()[2]) == "critical" 

    @mock.patch('core.input.register')
    def test_register_left_mouse_action(self, input_register_mock):
        module = build_module()

        input_register_mock.assert_called_with(
            module,
            button=core.input.LEFT_MOUSE,
            cmd='gnome-system-monitor'
        )





# pylint: disable=C0111,R0903

"""Displays the current date and time in Persian(Jalali) Calendar.

Requires the following python packages:
    * jdatetime

Parameters:
    * datetime.format: strftime()-compatible formatting string. default: "%A %d %B" e.g., "جمعه ۱۳ اسفند"
    * datetime.locale: locale to use. default: "fa_IR"
"""

import jdatetime

import core.decorators
from modules.core.datetime import Module as dtmodule


class Module(dtmodule):
    @core.decorators.every(minutes=1)
    def __init__(self, config, theme):
        super().__init__(config, theme, dtlibrary=jdatetime)

    def default_format(self):
        return "%A %d %B"

    def default_locale(self):
        return ("fa_IR", "UTF-8")


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import pytest
import json

import util.location


@pytest.fixture
def urllib_req(mocker):
    util.location.reset()
    return mocker.patch("util.location.urllib.request")


@pytest.fixture
def secondaryLocation():
    return {
        "country": "Middle Earth",
        "longitude": "10.0",
        "latitude": "20.5",
        "ip": "127.0.0.1",
    }


@pytest.fixture
def primaryLocation():
    return {
        "country_name": "Rivia",
        "longitude": "-10.0",
        "latitude": "-23",
        "ip": "127.0.0.6",
    }


def test_primary_provider(urllib_req, primaryLocation):
    urllib_req.urlopen.return_value.read.return_value = json.dumps(primaryLocation)

    assert util.location.country() == primaryLocation["country_name"]
    assert util.location.coordinates() == (
        primaryLocation["latitude"],
        primaryLocation["longitude"],
    )
    assert util.location.public_ip() == primaryLocation["ip"]


def test_secondary_provider(mocker, urllib_req, secondaryLocation):
    urlopen = mocker.MagicMock()
    urlopen.read.return_value = json.dumps(secondaryLocation)
    urllib_req.urlopen.side_effect = [RuntimeError(), urlopen]

    assert util.location.country() == secondaryLocation["country"]
    assert util.location.coordinates() == (
        secondaryLocation["latitude"],
        secondaryLocation["longitude"],
    )
    assert util.location.public_ip() == secondaryLocation["ip"]


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays the name, IP address(es) and status of each available network interface.

Requires the following python module:
    * netifaces

Requires the following executable:
    * iw
    * (until and including 2.0.5: iwgetid)

Parameters:
    * nic.exclude: Comma-separated list of interface prefixes (supporting regular expressions) to exclude (defaults to 'lo,virbr,docker,vboxnet,veth,br,.*:avahi')
    * nic.include: Comma-separated list of interfaces to include
    * nic.states: Comma-separated list of states to show (prefix with '^' to invert - i.e. ^down -> show all devices that are not in state down)
    * nic.format: Format string (defaults to '{intf} {state} {ip} {ssid} {strength}')
    * nic.strength_warning: Integer to set the threshold for warning state (defaults to 50)
    * nic.strength_critical: Integer to set the threshold for critical state (defaults to 30)
"""

import re
import shutil
import netifaces
import subprocess

import core.module
import core.decorators
import util.cli
import util.format


class Module(core.module.Module):
    @core.decorators.every(seconds=5)
    def __init__(self, config, theme):
        widgets = []
        super().__init__(config, theme, widgets)
        self._exclude = util.format.aslist(
            self.parameter("exclude", "lo,virbr,docker,vboxnet,veth,br,.*:avahi")
        )
        self._include = util.format.aslist(self.parameter("include", ""))

        self._states = {"include": [], "exclude": []}
        for state in tuple(
            filter(len, util.format.aslist(self.parameter("states", "")))
        ):
            if state[0] == "^":
                self._states["exclude"].append(state[1:])
            else:
                self._states["include"].append(state)
        self._format = self.parameter("format", "{intf} {state} {ip} {ssid} {strength}")

        self._strength_threshold_critical = self.parameter("strength_critical", 30)
        self._strength_threshold_warning = self.parameter("strength_warning", 50)

        # Limits for the accepted dBm values of wifi strength
        self.__strength_dbm_lower_bound = -110
        self.__strength_dbm_upper_bound = -30

        self.iw = shutil.which("iw")
        self._update_widgets(widgets)

    def update(self):
        self._update_widgets(self.widgets())

    def state(self, widget):
        states = []

        if widget.get("state") == "down":
            states.append("critical")
        elif widget.get("state") != "up":
            states.append("warning")

        intf = widget.get("intf")
        iftype = "wireless" if self._iswlan(intf) else "wired"
        iftype = "tunnel" if self._istunnel(intf) else iftype

        # "strength" is none if interface type is not wlan
        strength = widget.get("strength")
        if self._iswlan(intf) and strength:
            if strength < self._strength_threshold_critical:
                states.append("critical")
            elif strength < self._strength_threshold_warning:
                states.append("warning")

        states.append("{}-{}".format(iftype, widget.get("state")))

        return states

    def _iswlan(self, intf):
        # wifi, wlan, wlp, seems to work for me
        if intf.startswith("w"):
            return True
        return False

    def _istunnel(self, intf):
        return intf.startswith("tun") or intf.startswith("wg")

    def get_addresses(self, intf):
        retval = []
        try:
            for ip in netifaces.ifaddresses(intf).get(netifaces.AF_INET, []):
                if ip.get("addr", "") != "":
                    retval.append(ip.get("addr"))
        except Exception:
            return []
        return retval

    def _excluded(self, intf):
        for e in self._exclude:
            if re.match(e, intf):
                return True
        return False

    def _update_widgets(self, widgets):
        self.clear_widgets()
        interfaces = []
        for i in netifaces.interfaces():
            if not self._excluded(i):
                interfaces.append(i)
        interfaces.extend([i for i in netifaces.interfaces() if i in self._include])

        for intf in interfaces:
            addr = []
            state = "down"
            for ip in self.get_addresses(intf):
                addr.append(ip)
                state = "up"

            if len(self._states["exclude"]) > 0 and state in self._states["exclude"]:
                continue
            if (
                len(self._states["include"]) > 0
                and state not in self._states["include"]
            ):
                continue

            strength_dbm = self.get_strength_dbm(intf)
            strength_percent = self.convert_strength_dbm_percent(strength_dbm)

            widget = self.widget(intf)
            if not widget:
                widget = self.add_widget(name=intf)
            # join/split is used to get rid of multiple whitespaces (in case SSID is not available, for instance
            widget.full_text(
                " ".join(
                    self._format.format(
                        ip=", ".join(addr),
                        intf=intf,
                        state=state,
                        strength=str(strength_percent) + "%" if strength_percent else "",
                        ssid=self.get_ssid(intf),
                    ).split()
                )
            )
            widget.set("intf", intf)
            widget.set("state", state)
            widget.set("strength", strength_percent)

    def get_ssid(self, intf):
        if not self._iswlan(intf) or self._istunnel(intf) or not self.iw:
            return ""

        iw_info = util.cli.execute("{} dev {} info".format(self.iw, intf))
        for line in iw_info.split("\n"):
            match = re.match(r"^\s+ssid\s(.+)$", line)
            if match:
                return match.group(1)

        return ""

    def get_strength_dbm(self, intf):
        if not self._iswlan(intf) or self._istunnel(intf) or not self.iw:
            return None

        with open("/proc/net/wireless", "r") as file:
            for line in file:
                if intf in line:
                    # Remove trailing . by slicing it off ;)
                    strength_dbm = line.split()[3][:-1]
                    return util.format.asint(strength_dbm,
                                minimum=self.__strength_dbm_lower_bound,
                                maximum=self.__strength_dbm_upper_bound)

        return None

    def convert_strength_dbm_percent(self, signal):
        return int(100 * ((signal + 100) / 70.0)) if signal else None


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Shows a widget for each connected screen and allows the user to enable/disable screens.

Parameters:
    * xrandr.overwrite_i3config: If set to 'true', this module assembles a new i3 config
      every time a screen is enabled or disabled by taking the file '~/.config/i3/config.template'
      and appending a file '~/.config/i3/config.<screen name>' for every screen.
    * xrandr.autoupdate: If set to 'false', does *not* invoke xrandr automatically. Instead, the
      module will only refresh when displays are enabled or disabled (defaults to true)
    * xrandr.exclude: Comma-separated list of display name prefixes to exclude
    * xrandr.autotoggle: Boolean flag to automatically enable new displays (defaults to false)
    * xrandr.autotoggle_side: Which side to put autotoggled displays on ('right' or 'left', defaults to 'right')

Requires the following python module:
    * (optional) i3 - if present, the need for updating the widget list is auto-detected

Requires the following executable:
    * xrandr
"""

import re
import sys

import core.module
import core.input
import core.decorators

from bumblebee_status.discover import utility

import util.cli
import util.format

try:
    import i3
except Exception:
    pass


RESOLUTION_REGEX = re.compile(r"\d+x\d+\+(\d+)\+\d+")


class DisplayInfo:
    def __init__(self, name, resolution, connected, added, removed):
        self.name = name
        self.active = resolution is not None
        self.connected = connected
        self.added = added
        self.removed = removed

        self.position = int(resolution.group(1)) if self.active else sys.maxsize
        self.state = "on" if self.active else "off"

    def __str__(self):
        return "DisplayInfo(name={}, active={}, connected={}, added={}, removed={}, position={}, state={})".format(
            self.name,
            self.active,
            self.connected,
            self.added,
            self.removed,
            self.position,
            self.state,
        )

    def __repr__(self):
        return str(self)


class Module(core.module.Module):
    @core.decorators.every(seconds=5)  # takes up to 5s to detect a new screen
    def __init__(self, config, theme):
        super().__init__(config, theme, [])

        self._exclude = tuple(util.format.aslist(self.parameter("exclude")))
        self._autoupdate = util.format.asbool(self.parameter("autoupdate", True))
        self._autotoggle = util.format.asbool(self.parameter("autotoggle", False))
        self._autotoggle_side = self.parameter("autotoggle_side", "right")

        self._connected_displays = []
        self._active_displays = []
        self._initialized = False

        try:
            i3.Subscription(self._output_update, "output")
        except Exception:
            pass

    def _output_update(self, *_):
        self.update(force=True)

    def _query_displays(self):
        displays = []

        for line in util.cli.execute("xrandr -q").split("\n"):
            # disconnected or connected
            if "connected" not in line:
                continue

            name = line.split(" ", 2)[0]
            resolution = RESOLUTION_REGEX.search(line)
            active = resolution is not None

            connected = "disconnected" not in line
            added = connected and not active and name not in self._connected_displays
            removed = not connected and active and name in self._active_displays

            displays.append(DisplayInfo(name, resolution, connected, added, removed))

        self._connected_displays = [
            display.name for display in displays if display.connected
        ]
        self._active_displays = [display.name for display in displays if display.active]

        return displays

    def update(self, force=False):
        if not (self._autoupdate or force or not self._initialized):
            return

        self.clear_widgets()

        for display in self._query_displays():
            if display.name.startswith(self._exclude):
                continue

            if self._initialized and self._autotoggle:
                if display.added:
                    self._enable_display(display.name, self._autotoggle_side)
                elif display.removed:
                    self._disable_display(display.name)

            if not display.connected:
                continue

            widget = self.add_widget(full_text=display.name, name=display.name)
            core.input.register(widget, button=1, cmd=self._toggle)
            core.input.register(widget, button=3, cmd=self._toggle)

            widget.set("state", display.state)
            widget.set("pos", display.position)

        if not self._autoupdate:
            widget = self.add_widget(full_text="")
            widget.set("state", "refresh")
            core.input.register(widget, button=1, cmd=self.update)

        self._initialized = True

    def state(self, widget):
        return widget.get("state", "off")

    def _toggle_cmd(self):
        if util.format.asbool(self.parameter("overwrite_i3config", False)):
            return utility("toggle-display.sh")
        else:
            return "xrandr"

    def _disable_display(self, name):
        if len(self._active_displays) > 1:
            util.cli.execute("{} --output {} --off".format(self._toggle_cmd(), name))

    def _enable_display(self, name, side=None):
        # TODO: is there ever a case when there isn't a neighbor?
        command = "{} --output {} --auto".format(self._toggle_cmd(), name)
        if side and self._active_displays:
            neighbor_index = 0 if side == "left" else -1
            command += " --{}-of {}".format(side, self._active_displays[neighbor_index])

        util.cli.execute(command)

    def _toggle(self, event):
        widget = self.widget(widget_id=event["instance"])

        if widget.get("state") == "on":
            self._disable_display(widget.name)
        else:
            side = "left" if event["button"] == core.input.LEFT_MOUSE else "right"
            self._enable_display(widget.name, side)

        self.update(force=True)


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import pytest

pytest.importorskip("feedparser")

pytest.importorskip("webbrowser")

pytest.importorskip("tempfile")

pytest.importorskip("random")

def test_load_module():
    __import__("modules.contrib.rss")




import pytest
from unittest import TestCase, mock
from freezegun import freeze_time

import core.config
import core.input
import core.widget
import modules.core.date

pytest.importorskip("datetime")

def build_module(args = []):
    config = core.config.Config(args)
    return modules.core.date.Module(config=config, theme=None)

def build_widget():
    return core.widget.Widget()

class DateTest(TestCase):
    def setup_class(self):
        locale_patcher = mock.patch('locale.getdefaultlocale')
        locale_mock = locale_patcher.start()
        locale_mock.return_value = ('en-US', 'UTF-8')

        self.widget = build_widget()

    def test_load_module(self):
        __import__("modules.core.date")

    @freeze_time('2020-10-15 03:25:59')
    def test_default_format(self):
        module = build_module()
        assert module.full_text(self.widget) == '10/15/2020'

    @freeze_time('2020-10-20 12:30:00')
    def test_custom_format(self):
        module = build_module(['-p', 'date.format=%d.%m.%y'])
        assert module.full_text(self.widget) == '20.10.20'

    @freeze_time('2020-01-10 10:20:30')
    @mock.patch('locale.getdefaultlocale')
    def test_invalid_locale(self, locale_mock):
        locale_mock.return_value = ('in-IN', 'UTF-0')

        module = build_module()
        assert module.full_text(self.widget) == '01/10/2020'

    @mock.patch('core.input.register')
    def test_register_left_mouse_action(self, input_register_mock):
        module = build_module()

        input_register_mock.assert_called_with(
            module,
            button=core.input.LEFT_MOUSE,
            cmd='calendar'
        )





# pylint: disable=C0111,R0903

"""Displays CPU utilization across all CPUs.

By default, opens `gnome-system-monitor` on left mouse click.

Requirements:
    * the psutil Python module for the first three items from the list above
    * gnome-system-monitor for default mouse click action

Parameters:
    * cpu.warning : Warning threshold in % of CPU usage (defaults to 70%)
    * cpu.critical: Critical threshold in % of CPU usage (defaults to 80%)
    * cpu.format  : Format string (defaults to '{:.01f}%')
    * cpu.percpu  : If set to true, show each individual cpu (defaults to false)
"""

import psutil

import core.module
import core.widget
import core.input

import util.format


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, [])
        self._percpu = util.format.asbool(self.parameter("percpu", False))

        for idx, cpu_perc in enumerate(self.cpu_utilization()):
            widget = self.add_widget(name="cpu#{}".format(idx), full_text=self.utilization)
            widget.set("utilization", cpu_perc)
            widget.set("theme.minwidth", self._format.format(100.0 - 10e-20))

        core.input.register(
            self, button=core.input.LEFT_MOUSE, cmd="gnome-system-monitor"
        )

    @property
    def _format(self):
        return self.parameter("format", "{:.01f}%")

    def utilization(self, widget):
        return self._format.format(widget.get("utilization", 0.0))

    def cpu_utilization(self):
        tmp = psutil.cpu_percent(percpu=self._percpu)
        return tmp if self._percpu else [tmp]

    def update(self):
        for idx, cpu_perc in enumerate(self.cpu_utilization()):
            self.widgets()[idx].set("utilization", cpu_perc)

    def state(self, widget):
        return self.threshold_state(widget.get("utilization", 0.0), 70, 80)


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
